# Author: Bryan C. Ward (bryan.ward@bucknell.edu)
# Last Modified: Mon Jan 10, 2011 at 04:15 PM -0500

from safe.helpers import TupleIndexer
from safe.boolean import *
from copy import deepcopy

class LinearDesignPointGenerator(object):
    """ A Design Point Generation algorithm.

    This algorithm iterates through the size of the parameter space checking if
    each point is in the experimental design.  It conforms to the
    specifications required by the Experiment class so that it can be used to
    generate the design points for the experiment.

    """

    def __init__(self, fal, restriction_expression):
        """ Constructor.

        fal is the factors and levels dictionary.
        restrictions is the list of invalid factor level tuples which are invalid.

        """

        self.points = []
        self.restriction_expression = restriction_expression
        self.fal = fal
        self.tuplecount = 1
        self.master_index = 0
        divisor = 1
        self.indexers = {}

        # from the python documentation:
        #
        # The keys() method of a dictionary object returns a list of all
        # the keys used in the dictionary, in arbitrary order (if you want
        # it sorted, just apply the sorted() function to it).
        #
        # So I changed it to sorted() everywhere.
        # - Bryan
        
        for factor in sorted(self.fal.keys()):
            size = len(self.fal[factor])
            self.indexers[factor] = TupleIndexer(divisor, size)
            # print('%-10s indexing function : (i/%s) %% %s' % (factor, divisor, size))
            divisor = divisor * size
            self.tuplecount = self.tuplecount * size

    def __iter__(self):
        """ Make this class an iterator. """
        return self

    # Delete this later.
    def set_restriction_expression(self, rexp):
        self.restriction_expression = rexp

    def next(self):
        """ Get the next design point in the iterator. """
        while True: # necessary if self.master_index is invalid
            if self.master_index == self.tuplecount:
                raise StopIteration
            else:
                point = {}
                for factor in sorted(self.fal.keys()):
                    index = self.indexers[factor].index(self.master_index)
                    point[factor] = self.fal[factor][index]
                self.master_index += 1
                if self.valid(point):
                    return point

    def valid(self, point):
        # By default, assume this is a valid point.  If there is no
        # restriction_expression at all, then it must be valid.
        is_valid = True
        if self.restriction_expression != None:
            is_valid = self.restriction_expression.evaluate(point)

        return is_valid

    
    def design_point_by_index(self, index):
        point = {}
        for factor in sorted(self.fal.keys()):
            i = self.indexers[factor].index(index)
            point[factor] = self.fal[factor][i]
        return point

class BacktrackingDesignPointGenerator(object):
    """ Another design point generation algorithm.

    This algorithm takes an alternative approach to design point generation
    using a backtracking algorithm.  A level is applied to a factor and then
    the list of available levels for each remaining factor are updated.  The
    next most constrained factor chooses a level, and recurs down until all
    factors have had levels applied.  At this point a design point has been
    found and returned.  Such an algorithm will excel in cases in which the
    original cartesian product of the parameter space is larger much larger
    than the number of design points in the experiment.

    """

    def __init__(self, fal, restrictions):
        self.fal = fal
        self.restrictions = restrictions
        self.stack = [(sorted(self.fal.keys()), {},
            deepcopy(self.fal))]

    def __iter__(self):
        """ Make this class an iterator. """
        return self

    def new_constraints(self, factor, level, myFactorLevels):
        """Compute all of the new constraints after assigning factor a
        level.

        TODO: change this method to reflect dictionaries of restrictions
        instead of only 2-tuples.

        """

        constraints = []
        #for r in self.restrictions:
        #    if r[0][0] == factor and r[0][1] == level and\
        #        r[1][1] in myFactorLevels[r[1][0]]:

        #        constraints.append(r[1])
        #    elif r[1][0] == factor and r[1][1] == level and\
        #        r[0][1] in myFactorLevels[r[0][0]]:

        #        constraints.append(r[0])
        return constraints

    def next(self):
        """ Get the next design point in the iterator. """

        while len(self.stack) > 0:
            keys, sim, myFactorLevels= self.stack.pop()

            if len(keys) == 0:
                return sim
            else:
                constraints = sorted(
                        [(len(myFactorLevels[x]),x) for x in keys])
                count, factor = constraints.pop(0)
                keys.remove(factor)

                for level in myFactorLevels[factor]:
                    sim[factor] = level
                    new_cts = self.new_constraints(
                            factor, level, myFactorLevels) 

                    for f,l in new_cts:
                        myFactorLevels[f].remove(l)

                    self.stack.append((deepcopy(keys), deepcopy(sim),
                        deepcopy(myFactorLevels)))

                    for f,l in new_cts:
                        myFactorLevels[f].append(l)

        # if we fall through the while loop, we are done
        raise StopIteration
