#!/usr/bin/env python

from twisted.python import components, log
from twisted.cred import portal, checkers, credentials
from twisted.conch import error, avatar, recvline, interfaces as \
conchinterfaces
from twisted.conch.ssh import channel, factory, userauth, connection, \
session, common, filetransfer
from twisted.conch.insults import insults
from twisted.conch import unix
from twisted.application import service, internet
from twisted.protocols import ftp
from zope.interface import implements

from twisted.conch.ssh.keys import Key
from fabric import network, tasks, colors

import sys
import os
import hashlib
import ConfigParser
import tarfile
import subprocess
import fabfile

from django.conf import settings
#Uses same settings as django
os.environ.setdefault("DJANGO_SETTINGS_MODULE","web.source.settings")

from web.safe.models import *

class FileTransferAvatar(avatar.ConchUser):
    """ Avatar that implements the file transfer protocol.

    Commands are run as SAFE user so they need to be strictly
    sanitized so that the user stays within his/her compartment.

    """

    def __init__(self, username):
        avatar.ConchUser.__init__(self)
        self.username = username
        self.channelLookup['session'] = session.SSHSession
        self.subsystemLookup['sftp'] = filetransfer.FileTransferServer
        #Check if user directory exists
        if not os.path.exists(self.getHomeDir()):
          #If it does not exist, create a new one
          os.makedirs(self.getHomeDir())

    def _runAsUser(self, f, *args, **kw):
        try:
            f = iter(f)
        except TypeError:
            f = [(f, args, kw)]
        for i in f:
            func = i[0]
            args = len(i) > 1 and i[1] or ()
            kw = len(i) > 2 and i[2] or {}
            # Make sure the user stays within their compartment.
            # Do this by making sure any directory is prefixed
            # with their compartment path.
            compartment = os.path.join(os.getcwd(), 'users', self.username)
            goodPath = True
            
            #Check for the first argument if it does not starts with another path that does not belong to the user
            arg=args[0]
            # TODO: Possibly match against regex instead of all 
            # strings (although this should work)
            if (isinstance(arg, str) and not arg.startswith(compartment)):
                # Return permission denied error message.
                goodPath = False
                raise filetransfer.SFTPError(filetransfer.FX_PERMISSION_DENIED, "Permission Denied")
                    
            if goodPath:
                r = func(*args, **kw)
        
        return r

    def getHomeDir(self):
        return os.path.join(os.getcwd(), 'users', self.username)

    def attemptExpExecution(self):
         
        def unpack(archivePath, basePath):
            """ Helper method to unpack the archive recieved from the user. 

            This method uses Python's tarfile library and is system 
            independent.

            """

            resolved = lambda x: os.path.realpath(os.path.abspath(x))

            def _badpath(path, base):
                """ Make sure that paths do not escape compartment. 

                Join will ignore base if path is absolute.
                
                """

                return not resolved(os.path.join(base,path)).startswith(base)

            def _badlink(info, base):
                """ Make sure links do not exist to files/dirs outside compartment. 

                Links are interpreted relative to the directory containing the link
                
                """

                tip = resolved(os.path.join(base, os.path.dirname(info.name)))
                return _badpath(info.linkname, base=tip)

            def safemembers(members):
                """ Return a list of safe members to extract from a given tarfile."""

                base = resolved(".")

                for finfo in members:
                    if _badpath(finfo.name, base):
                        print finfo.name, "is blocked (illegal path)"
                    elif finfo.issym() and _badlink(finfo, base):
                        print finfo.name, "is blocked: Hard link to", finfo.linkname
                    elif finfo.islnk() and _badlink(finfo, base):
                        print finfo.name, "is blocked: Symlink to", finfo.linkname
                    else:
                        yield finfo
            
            if os.path.exists(archivePath):
                print(colors.green("SAFE: Archive found, extracting contents."))
                tar = tarfile.open(archivePath)
                tar.extractall(basePath, members=safemembers(tar))
                tar.close()
            else:
                print(colors.red("SAFE: Archive not found, checking for already filled compartment.\n"))

        def configure(configPath, basePath):
            """ Reads in and parses the experiment configuration in the user's repository. """

            def parseConfig(configPath, basePath):
                """ Parses the experiment configuration. 
                
                It returns a 3-tuple of the experiment name, the path to the 
                experiment XML configuration, and the path to the experiment 
                NS-3 script.

                """

                config = ConfigParser.ConfigParser()
                config.read(configPath)
                name = config.get('Experiment', 'name')
                xml = os.path.join(basePath, os.path.normpath(config.get('Experiment', 'descriptionXML')))
                script = config.get('Experiment', 'script')
               
                return (name, xml, script)
            
            if os.path.exists(configPath):
                print(colors.green("SAFE: Configuration found, attempting to parse."))
                return parseConfig(configPath, basePath)
            else:
                print(colors.red("SAFE: No experiment configuration found, no experiment executed."))

        def execute(self, expName, expDescription, expScript):
            """ Distribute, unpack, build, and execute experiment on all remote clients. 
            
            See fabfile.py for how this is accomplished.

            """

            descServerPath = os.path.join(os.getcwd(), 'users', self.username, expDescription)
            if os.path.exists(descServerPath):
                print(colors.green("SAFE: Description found, executing %s's experiment." % self.username))
                # Possible race condition between start of EEM and execution of clients. Since EEM
                # is started before distribution, it is very likely that this will never happen.
                # The EEM has a very small startup time, after which it is listening on the specified
                # port.
                tasks.execute(fabfile.startEEM, self.username, expName, expScript, descServerPath)
                tasks.execute(fabfile.distribute, self.username)
                tasks.execute(fabfile.prepare)
                tasks.execute(fabfile.startClients)
            else:
                print(colors.red("SAFE: No experiment description found, no experiment executed."))
        
        try:
            # Initialize some useful path strings based on this Avatar
            basePath = os.path.join(os.getcwd(), 'users', self.username)
            archivePath = os.path.join(basePath, 'in.tar.gz')
            configPath = os.path.join(os.getcwd(), 'users', self.username, "expconfig")

            unpack(archivePath, basePath)
            config = configure(configPath, basePath)
            if config is not None:
                execute(self, config[0], config[1], config[2])
        except:
            print "Unexpected error:", sys.exc_info()[0]
            raise  
        else:
            return 0
        finally:
            # Need to be sure that all clients get disconnected.
            network.disconnect_all()

if unix:
    class FileTransferForAvatar(unix.SFTPServerForUnixConchUser):

        def gotVersion(self, otherVersion, extData):
            return {}

        def extendedRequest(self, extName, extData):
            if extName == 'testExtendedRequest':
                return 'bar'
            raise NotImplementedError

    components.registerAdapter(FileTransferForAvatar, FileTransferAvatar, filetransfer.ISFTPServer)

class SFTPRealm():
    implements(portal.IRealm)
    """ This realm handles the granting of an sFTP session to a registered user. 
    
    If a user successfully authenticates, he/she will be granted an avatar that 
    can handle sFTP commands.

    """

    def requestAvatar(self, avatarID, mind, *interfaces):
        """ Return an Avatar that implements the correct protocol. 
        
        In this case this is just sFTP. When that Avatar disconnects, 
        attempt experiment execution.

        """

        a = FileTransferAvatar(avatarID)
        return interfaces[0], a, a.attemptExpExecution

def getRSAKeys():
    """ Generate RSA keys in case user has not done this already.  
    
    This method is not necessary, but may be helpful to some.

    """
    
    if not (os.path.exists('config/public.key') and os.path.exists('config/private.key')):
        # Generate a RSA keypair
        print "Generating RSA keypair..."
        from Crypto.PublicKey import RSA
        KEY_LENGTH = 1024
        rsaKey = Key(RSA.generate(KEY_LENGTH))
        publicKeyString = rsaKey.public().toString('OPENSSH')
        privateKeyString = rsaKey.toString('OPENSSH')
        
        # Save keys for next time
        file('config/public.key', 'w+b').write(publicKeyString)
        file('config/private.key', 'w+b').write(privateKeyString)
        print "done."
    else:
        publicKeyString = file('config/public.key').read()
        privateKeyString = file('config/private.key').read()
    return publicKeyString, privateKeyString

### DATABASE PASSWORD CHECKER FOR TWISTED ###
class DBCredentialsChecker(object):
  implements(checkers.ICredentialsChecker)
  credentialInterfaces = (credentials.IUsernamePassword,
      credentials.IUsernameHashedPassword)

  #TODO make it asynchronous
  def requestAvatarId(self, credentials):
    #Try to get corresponding user object
    try:
      user=User.objects.get(username=credentials.username)
      #Calculate Password hash
      password_hash = hashlib.sha256(credentials.password).hexdigest()
      #It is necessary to be admin to create new experiment
      if user.pass_hash == password_hash and user.is_admin:
        #Success login
        return credentials.username
      else:
        return defer.fail(error.UnauthorizedLogin("Incorrect password"))
    except User.DoesNotExist, User.MultipleObjectsReturned:
      return defer.fail(error.UnauthorizedLogin("Invalid user"))

if __name__ == "__main__":
    #log.startLogging(sys.stdout)
    sshFactory = factory.SSHFactory()
    sshFactory.portal = portal.Portal(SFTPRealm())
    
    # Check for Credential using database
    checker=DBCredentialsChecker()
    sshFactory.portal.registerChecker(checker)

    pubKeyString, privKeyString = getRSAKeys()
    sshFactory.publicKeys = {
        'ssh-rsa': Key.fromString(data=pubKeyString)}
    sshFactory.privateKeys = {
        'ssh-rsa': Key.fromString(data=privKeyString)}

    from twisted.internet import reactor
    reactor.listenTCP(2022, sshFactory)
    reactor.run()
