from twisted.trial import unittest
from twisted.test import proto_helpers
from twisted.internet.defer import Deferred, DeferredList
from struct import *
from yoyo.migrate.connections import parse_uri
from ConfigParser import ConfigParser
from random import randint

from safe.debug import *
from safe.manager import *
from safe.messages import *
from safe.models import Experiment, Template, Description

class ManagerTest(object):
    """Abstract class which provides setup and teardown.

    Because the ExperimentFactory and the ExperimentManager have similar setUp
    and tearDown methods, by using inheritence we can avoiding copy/paste.

    """

    def setUp(self):
        """Setup things before the testing"""

        def createExperiment(results):
            """Called when the config files have been saved."""

            self.name = self.mktemp()

            self.experiment = Experiment(self.dbpool,
                name = self.name,
                template_id = results[0][1],
                description_id = results[1][1])
            
            self.factory = ExperimentFactory(self.dbpool,
                experiment = self.experiment)

            self.protos = []

            for i in xrange(10):
                proto = self.factory.buildProtocol(('127.0.0.1',0))
                transport = proto_helpers.StringTransport()
                proto.makeConnection(transport)
                self.protos.append((proto,transport))

            self.proto = self.protos[0][0]
            self.transport = self.protos[0][1]
            
            self.experiment.prepare_for_execution().addCallback(answer.callback).addErrback(printError)

        answer = Deferred()

        config = ConfigParser()
        config.read("../config/server")
        dburi = config.get('TEST','dburi')

        _, user, password, host, port, database = parse_uri(dburi)

        self.dbpool = adbapi.ConnectionPool("psycopg2",
                host = host,
                database = database,
                port = port or 5432,
                password = password or "",
                user = user)
        
        self.template = Template(self.dbpool,
                template = open("../examples/template.xml").read())
        self.description = Description(self.dbpool,
                description = open("../parsing/CsmaExperiment.xml").read())
#                description = open("../examples/description.xml").read())

        dlist = DeferredList([
                self.template.save(),
                self.description.save()],
                fireOnOneErrback = True)

        dlist.addCallback(createExperiment)
        dlist.addErrback(printError)

        return answer

    def tearDown(self):
        return DeferredList([
            self.experiment.delete(cascade = True),
            self.template.delete(cascade = True),
            self.description.delete(cascade = True)])

class ExperimentFactoryTestCase(ManagerTest, unittest.TestCase):
    """Tests for the factory which manages all of the connections to the
    individual clients.

    """

    #def registerProtos(self):
    #    for id,p in enumerate(self.protos):
    #        p.dataReceived(Register(str(id)).encode())
    #
    #def testProtocolsStored(self):
    #    """In setUp 10 protocols are initalized, test they are stored"""

    #    self.assertEqual(len(self.factory.protocols),10)

    #def testHostRegistration(self):
    #    """Test the factory keeps a log of which protocols are working on which
    #    simulations.
    #    
    #    """

    #    self.factory.register_host(None,5,self.proto)
    #    self.assertTrue(self.proto in self.factory.simulations[5])

    #    self.factory.register_host(5,6,self.proto)
    #    self.assertTrue(self.proto in self.factory.simulations[6])
    #    self.assertFalse(self.proto in self.factory.simulations[5])

    def testFinishedNotification(self):
        """Test that only the appropriate protocols get notifed when a
        simulation has finished.

        """

        #set up which simulations are active
        self.factory.register_host(None,5,self.protos[0][0])
        self.factory.register_host(None,5,self.protos[1][0])
        
        self.factory.notify_finished(5)

        #check only the appropriate protos get notified
        self.assertEqual(self.protos[0][1].value(),Finished().encode())
        self.assertEqual(self.protos[1][1].value(),Finished().encode())
        self.assertEqual(self.protos[2][1].value(),"")

    #def testTerminate(self):
    #    """Test that all protocols are terminated."""

    #    def validateAction(_):
    #        """Validate that the messages were sent and db updated."""

    #        retval = True
    #        for p in self.protos:
    #            if p[1].value() != Terminate().encode():
    #                retval = False

    #        answer.callback(True)

    #    answer = defer.Deferred()
    #    answer.addCallback(self.assertNotEqual,[(None,)])

    #    d = self.factory.terminate_all(finished=True)
    #    d.addCallback(validateAction)

    #    answer.addCallback(self.assertEqual,True)
    #    return answer

class TestExperimentManager(ManagerTest, unittest.TestCase):

    def testFragment(self):
        """Test the system doesn't do anything when a fragment is received"""

        self.proto.dataReceived(pack('!I',SAMPLE))
        self.assertEquals(self.transport.value(),"")

    def testHandleRegister(self):
        """Test the simulation client registration procedures."""

        answer = defer.Deferred()

        def check(saved):
            Host.find(self.dbpool, kernel = kernel).addCallback(
                lambda host: host.kernel.addCallback(answer.callback))

        kernel = self.mktemp()[-60:]
        if len(kernel) > 60:
            kernel = kernel[-60:]

        saving = self.proto._handle_register(Register(kernel))
        saving.addCallback(check)

        answer.addCallback(self.assertEqual, kernel)
        return answer

    def testHandleNextRequest(self):
        """Test the response for a request for another simulation."""

        kernel = self.mktemp()[-60:]
        if len(kernel) > 60:
            kernel = kernel[-60:]

        self.proto._handle_register(Register(kernel))

        d = self.proto._handle_next_request(NextRequest())
        d.addCallback(lambda _: self.proto.design_point)
        d.addCallback(self.assertIsInstance,DesignPoint)

        return d

    def testNextReply(self):
        msg = NextReply("<testing/>")
        self.assertRaises(UnknownMessageError,
                self.proto.dataReceived,msg.encode())

##    def testSample(self):
##        msg = Sample(1.23,"node1","e2e",5.67)
##        self.proto.dataReceived(msg.encode())
##actually test things here
#
    def testFinished(self):
        msg = Finished()
        self.assertRaises(UnknownMessageError,
                self.proto.dataReceived,msg.encode())

    def testSendFinished(self):
        self.proto.send_finished()
        self.assertEqual(self.transport.value(),Finished().encode())

    def testSendTerminate(self):
        self.proto.send_terminate()
        self.assertEqual(self.transport.value(),Terminate().encode())

