from twisted.trial import unittest
from twisted.internet.defer import Deferred, DeferredList
from yoyo.migrate.connections import parse_uri
from ConfigParser import ConfigParser
import string
import random

from safe.exceptions import DatabaseError
from safe.models import *

class DatabaseModelTestCase(unittest.TestCase):
    """This TestCase tests the functionality of the DatabaseModel class.

    This test case uses the Experiment class to test the functionality of the
    DatabaseModel class it subclasses.  Experiment specific tests can be found
    in ExperimentTestCase

    """

    def setUp(self):
        """Setup things before the testing"""

        config = ConfigParser()
        config.read("../config/server")
        dburi = config.get('TEST','dburi')

        _, user, password, host, port, database = parse_uri(dburi)

        self.dbpool = adbapi.ConnectionPool("psycopg2",
                host = host,
                database = database,
                port = port or 5432,
                password = password or "",
                user = user)

        self.name = self.mktemp()
        self.experiment = Experiment(self.dbpool, name = self.name)

    def testSave(self):
        """Test the creation of an experiment."""

        return self.experiment.save().addCallback(self.assertTrue)

    def testGetAttribute(self):
        """Test that column can be queried for with attribute syntax."""

        def respond(results, name):
            self.experiment.name.addCallback(answer.callback)

        answer = Deferred()

        self.experiment.save().addCallback(respond, self.name)

        answer.addCallback(self.assertEqual,self.name)

        return answer

# this needs to be updated now that __getitem__ is just the cached value.
#    def testGetItem(self):
#        """Test that column can be queried for with dict syntax."""
#
#        def respond(results):
#            self.experiment['name'].addCallback(answer.callback)
#
#        answer = Deferred()
#
#        self.experiment.save().addCallback(respond)
#        
#        answer.addCallback(self.assertEqual,self.name)
#        return answer

    def testSetAttrColumn(self):
        """Test setting an attribute of a column name."""

        def respond(id, newname):
            self.experiment.name = newname
            self.experiment.name.addCallback(answer.callback)

        answer = Deferred()

        newname = self.mktemp()
        self.experiment.save().addCallback(respond, newname)

        answer.addCallback(self.assertEqual,newname)

        return answer
    
    def testUpdateColumn(self):
        """Test setting an attribute of a column name."""

        def respond(id, newname):
            self.experiment.name = newname
            self.experiment.update().addCallback(answer.callback)

        answer = Deferred()
        newname = self.mktemp()

        self.experiment.save().addCallback(respond, newname)

        answer.addCallback(self.assertEqual,None)
        return answer

    def testSetItemColumn(self):
        """Test setting a column value via dict syntax."""

        def respond(id, newname):
            self.experiment.name = newname
            self.experiment.name.addCallback(answer.callback)

        answer = Deferred()
        newname = self.mktemp()

        self.experiment.save().addCallback(respond, newname)

        answer.addCallback(self.assertEqual,newname)
        return answer

    def testDelete(self):
        """Test that experiments are removed from the database with delete."""

        def respond(id):
            e.delete().addCallback(answer.callback)

        answer = Deferred()
        name = self.mktemp()
        e = Experiment(self.dbpool, name = name)
        e.save().addCallback(respond)

        answer.addCallback(self.assertEqual, None)
        return answer

    def testCachedAttribute(self):
        """Test that a cached value is returned.

        The initial value given when the Experiment object is constructud
        should act as a cached attribute that has not yet been stored in the
        database.

        """

        self.experiment.name.addCallback(self.assertEqual,self.name)

    def testFindNone(self):
        """Test that find returns None when there are no rows."""

        return Experiment.find(self.dbpool, id=-1).addCallback(
                self.assertEqual,None)

    def testFindOne(self):
        """Test that find returns a single object when finding one."""

        def respond(id):
            Experiment.find(self.dbpool, id = id).addCallback(answer.callback)

        answer = Deferred()

        self.experiment.save().addCallback(respond)

        answer.addCallback(self.assertIsInstance,Experiment)
        return answer

    def tearDown(self):
        self.experiment.delete()
