from twisted.python import util
from twisted.internet.defer import Deferred, DeferredList
from twisted.application import service

from zope.interface import Interface, implements

from nevow import rend, static, url, athena
from nevow import tags as t
from nevow.stan import directive
from nevow.loaders import xmlstr, stan

from formless import annotate, webform

from safe.webmanager.template import TemplatedPage, ITemplatedPage
from safe.webmanager.plotting import PlottingPage
from safe.webmanager.results import ResultsPage
from safe.webmanager.design_points import DesignPointsPage
from safe.models import Experiment
from safe.manager import ExecutionManagerService
from safe.debug import *

class NewExperimentPage(TemplatedPage):
    """A page which allows the user to create a new experiment.
    
    I haven't learned yet how to customize the generated form, but this is
    likely something that should be done eventually.  I tried making a form
    myself so I could customize things manually, but I never could get it to
    work right.
    
    """
    implements(ITemplatedPage)

    def __init__(self,dbpool):
        super(TemplatedPage,self).__init__()
        self.dbpool = dbpool

    def render_content(self, ctx, data):
        return ctx.tag[t.invisible(render=directive("experimentForm"))]

    def render_experimentForm(self, ctx, data):
        return ctx.tag.clear()[
                webform.renderForms()]

    def newExperiment(self, **experiment):
        """Method invoked when a new experiment form is submitted."""

        def saveExperiment(
                (
                    (template_status,(template_id,)),
                    (restrictions_status,(restrictions_id,)),
                    (descriptions_status,(descriptions_id,))
                )
            ):

            if template_status and\
                    restrictions_status and\
                    descriptions_status:

                query = """
                INSERT INTO experiments
                (name, created_at,
                    template_id, restrictions_id, descriptions_id) VALUES
                ($name$%s$name$,NOW(),%d,%d,%d) RETURNING id;
                """ % (experiment['name'],template_id[0],
                        restrictions_id[0],descriptions_id[0])

                self.dbpool.runQuery(query)
            else:
                raise DatabaseError(
                        "XML configuration saving failure.")

#$xml$ is a postgres mechanism to escape the xml string
        dlist = []
        query = """
        INSERT INTO templates (template) VALUES
        ($xml$%s$xml$) RETURNING id
        """ % experiment['template'].file.read()

        dlist.append(self.dbpool.runQuery(query))

        query = """
        INSERT INTO restrictions (restrictions) VALUES
        ($xml$%s$xml$) RETURNING id
        """ % experiment['restrictions'].file.read()
        
        dlist.append(self.dbpool.runQuery(query))
        
        query = """
        INSERT INTO descriptions (descriptions) VALUES
        ($xml$%s$xml$) RETURNING id
        """ % experiment['descriptions'].file.read()
        
        dlist.append(self.dbpool.runQuery(query))

        DeferredList(dlist).addCallback(saveExperiment)

    def bind_newExperiment(self, ctx):
        return [
                ('name', annotate.String(required=True)),
                ('descriptions', annotate.FileUpload(required=True)),
                ('restrictions', annotate.FileUpload(required=True)),
                ('template', annotate.FileUpload(required=True))
                ]

class ExperimentPage(TemplatedPage):
    """Resource responsible for showing information specific to a single
    experiment.
    
    """
    
    implements(ITemplatedPage)
    
    class Content(rend.Fragment):
        """Fragment to compute the content of the page."""

        docFactory = stan([
            t.invisible(data=directive("experiment"),
                render=directive("experiment"))[
                    t.p[
                        t.b["Experiment id: "],
                        t.slot(name="id")
                    ],
                    t.p[
                        t.b["Experiment name: "],
                        t.slot(name="name")
                    ],
                    t.p[
                        t.b["Experiment created at: "],
                        t.slot(name='created')
                    ],
                    t.p[
                        t.b["Experiment finished at: "],
                        t.slot(name='finished')
                    ],
            ]
        ])
        
        def __init__(self,experiment_id , dbpool, app):
            """Initalize the Experiment Content fragment with the db."""

            self.dbpool = dbpool
            self.experiment_id = experiment_id
            self.app = app

        def render_experiment(self, context, data):
            """Fill the slots with the computed data."""
            context.fillSlots('id',data[0])
            context.fillSlots('name',data[1])
            context.fillSlots('created',data[2])
            context.fillSlots('finished',data[3])
            return context

        def data_experiment(self,context,data):
            """Method to get the experiments in the database"""

            def sanitize(result):
                """Sanitize the database output for display in page."""
                if len(result) != 1:
                    answer.errback("Experiment Not Found")
                else:
                    fmt = "%m-%d-%Y %I:%M%p"
                    data = list(result[0])
                    for i in [2,3]:
                        if data[i]:
                            data[i] = data[i].strftime(fmt)
                        else:
                            data[i] = ""
                    answer.callback(data)

            query = """
            SELECT id, name, created_at, finished_at
            FROM experiments
            WHERE id = %d
            """ % (self.experiment_id)

            answer = Deferred()
            db = self.dbpool.runQuery(query)
            db.addCallback(sanitize)

            return answer

    def __init__(self,experiment,dbpool,app):
        """Initialize the experiment resource."""

        super(TemplatedPage, self).__init__()

        self.experiment = experiment
        self.experiment_id = self.experiment.cachedValue("id")
        self.dbpool = dbpool
        self.app = app
        
        self.children = {
            'design_points' : DesignPointsPage(experiment, self.dbpool),
            'results' : ResultsPage(self.dbpool,
                experiment = self.experiment),
            'plotting' : PlottingPage(self.experiment, self.dbpool)
        }

    def render_page_title(self,ctx,data):
        """Compute the title of the page."""

        def sanitize(result):
            """Prepare the output."""
            if len(result) != 1:
                answer.errback("Experiment not found")
            else:
                answer.callback("Experiment: %s" % result[0][0])

        query = """
        SELECT name FROM experiments WHERE id = %d
        """ % (self.experiment_id)
        
        answer = Deferred()
        dbd = self.dbpool.runQuery(query)
        dbd.addCallback(sanitize)

        return answer 

    def render_content(self,ctx,data):
        return ctx.tag[self.Content(self.experiment_id,self.dbpool,self.app)]

class StartExperimentElement(athena.LiveElement):
    jsClass = u'WebManager.StartButton'

    docFactory = stan(t.div(class_ = 'button',
            render=t.directive('liveElement'))[
        t.input(type="submit", value = "start",
            onclick = 'Nevow.Athena.Widget.get(this).start()')])

    def __init__(self, dbpool, experiment, app):
        self.experiment = experiment
        self.app = app
        self.dbpool = dbpool

    @athena.expose
    def start(self):
        """Method invoked when a stop button is pressed.

        This method is called by the browser with javascript.  Furthermore,
        this method is not called directly, but rather, called by __getattr__
        when it catches there is no handle_stop_8 method, and it instead calls
        handle_method(8).

        """

        def startService(status):
            myservice = ExecutionManagerService(self.dbpool,
                experiment = self.experiment)
            myservice.setName("Execution Manager %d" % id)
            myservice.setServiceParent(self.app)
            self.callRemote('startInfo', id)

        id = self.experiment.cachedValue("id")

        services = [srv.name for srv in service.IServiceCollection(self.app)]

        if "Execution Manager %d" % id not in services:
            self.experiment.prepare_for_execution().addCallback(startService)
        else:
            for srv in service.IServiceCollection(self.app):
                if srv.name == "Execution Manager %d" % id:
                    srv.startService()

class StopExperimentElement(athena.LiveElement):
    jsClass = u'WebManager.StopButton'

    def __init__(self, dbpool, experiment, app):
        self.dbpool = dbpool
        self.experiment = experiment
        self.app = app

        self.id = experiment.cachedValue("id")

        self.docFactory = stan(t.div(class_ = 'button', id = str(id),
            render=t.directive('liveElement'))[
        t.input(type="submit", value = "stop",
            onclick = 'Nevow.Athena.Widget.get(this).stop()')])
    
    @athena.expose
    def stop(self):
        """Method invoked when a start button is pressed.

        This method is called by the browser with javascript.  Furthermore,
        this method is not called directly, but rather, called by __getattr__
        when it catches there is no handle_stop_8 method, and it instead calls
        handle_method(8).

        """

        id = self.experiment.cachedValue("id")
        for srv in service.IServiceCollection(self.app):
            if srv.name == "Execution Manager %d" % id:
                srv.stopService()
                srv.factory.terminateAll(finished = False)
        self.callRemote('stopInfo', id)

class ExperimentsPage(TemplatedPage):
    """Resource responsible for showing the experiments in the database.

    This page also provides dynamic content which allows the user to start
    and stop execution managers.

    """

    addSlash = True

    page_title = "Experiment Manager"
    
    def __init__(self, dbpool, app):
        """Create the experiments resource."""
        super(TemplatedPage, self).__init__()
        self.dbpool = dbpool
        self.app = app

    def render_content(self, ctx, data):
        return ctx.tag[
            t.div(id='info')[""],
            t.table(data=directive("experiments"),
                render=directive("sequence"), border="1")[
                t.tr(pattern='header')[
                    t.th["Id"],
                    t.th["Name"],
                    t.th["Created"],
                    t.th["Finished"],
                    t.th["Start"],
                    t.th["Stop"],
                ],
                t.tr(pattern="item", render=directive("row"))[
                    t.td[t.slot(name="id")],
                    t.td[t.slot(name="name")],
                    t.td[t.slot(name="created")],
                    t.td[t.slot(name="finished")],
                    t.td[t.slot(name="start")],
                    t.td[t.slot(name="stop")],
                ]
            ],
            t.p[t.a(href="new")["New Experiment"]],
        ]

    def render_row(self, ctx, experiment):
        """Create each row of the table of experiments."""
        fmt = "%m-%d-%Y %I:%M%p"
        id = experiment.cachedValue('id')
        name = experiment.cachedValue('name')

        if experiment.cachedValue('created_at'):
            created = experiment.cachedValue('created_at').strftime(fmt)
        else:
            created = ""
        if experiment.cachedValue('finished_at'):
            finished = experiment.cachedValue('finished_at').strftime(fmt)
        else:
            finished = ""

        start = StartExperimentElement(self.dbpool, experiment, self.app)
        start.setFragmentParent(self)

        experiment.start_button = start

        stop = StopExperimentElement(self.dbpool, experiment, self.app)
        stop.setFragmentParent(self)

        experiment.stop_button = stop

        ctx.fillSlots('id', id)
        ctx.fillSlots('name',
                t.a(href='/experiments/%d/' % id)[name])
        ctx.fillSlots('created', created)
        ctx.fillSlots('finished', finished)
        ctx.fillSlots('start', start)
        ctx.fillSlots('stop', stop)
        return ctx
    
    def data_experiments(self, ctx, data):
        """Method to get the experiments in the database.

        TODO: Refactor to include improvements from safe.models.
        """

        return Experiment.find(self.dbpool,
                cache=['name', 'created_at',  'finished_at'])

    def checkActive(self, id):
        for srv in service.IServiceCollection(self.app):
            if "Execution Manager %d" % id == srv.name and srv.running:
                return True
        return False

    def childFactory(self, ctx, segment):
        if segment.isdigit():
            d = Experiment.find(self.dbpool, id=int(segment))
            d.addCallback(lambda e: ExperimentPage(e, self.dbpool, self.app))
            return d
        elif segment == "new":
            return NewExperimentPage(self.dbpool)
        else:
            from safe.webmanager.errors import The404Page
            return The404Page()
