from twisted.python import util
from zope.interface import Interface, implements
from nevow import rend, static
from nevow import tags as t, entities as e
from nevow.stan import directive
from nevow.loaders import xmlstr, stan
from nevow import athena

class ITemplatedPage(Interface):
    """
    An interface which ensures that all implementing classes override the
    render_content method which generates the content of the page.
    
    """

    def render_content(self,ctx,data):
        """Return the content to be placed in the global template."""

    def render_page_title(self,ctx,data):
        """Render the title of the page."""

class TemplatedPage(athena.LivePage):
    """A class containing the application template"""

    addSlash = True
    
    docFactory = stan(t.html[
        t.head[
            t.meta(name="keywords",
                content="simulation,automation,framework,experiments"),
            t.meta(name="description",
                content="simulation automation frame for experiments"),
            t.invisible(render=t.directive('liveglue')),
            t.title[t.invisible(render=directive("page_title"))],
            t.link(href="/style.css", rel='stylesheet', type='text/css'),
        ],
        t.body[
            t.div(id="container")[
                t.div(id="header")[
                    t.h1[t.a(href="/")
                        ["Simulation Automation Framework for Experiments"]]
                ],
                t.div(id="menu", render=directive("navigation_bar")),
                t.div(id="content")[
                    t.h2[t.invisible(render=directive("page_title"))],
                    t.invisible(render=directive("content"))
                ],
                t.div(id="footer")[
                    t.p[e.copy," 2010 Bucknell University"],
                    t.p["Developed by Bryan C. Ward"]
                ]
            ]
        ]
    ])

    page_title = "Simulation Automation Framework"
    scripts = ""

    def render_page_title(self, ctx, data):
        """Set the page title for the template.

        By placing this method in the TemplatedPage class, the title of the
        page can be set through an instance variable in any overriding classes.
        Also, the inherity classes may override this method for custom page
        titles.

        """

        return self.page_title

    def render_navigation_bar(self, ctx, data):
        """Create the navigation bar for the page.

        If the class has an experiment_id attribute, then the navigation bar
        should include links to the simulations, results and plotter.
        Otherwise, it should just be to the experiments and help pages.

        """

        def respondToID(id):
            experiment = "/experiments/%d/" % id
            dps = "/experiments/%d/design_points/" % id
            results = "/experiments/%d/results/" % id
            plotting = "/experiments/%d/plotting/" % id

            return ctx.tag[stan(t.ul[
                t.li[t.a(href=experiment)["Experiment"]],
                t.li[t.a(href=dps)["Design Points"]],
                t.li[t.a(href=results)["Results"]],
                t.li[t.a(href=plotting)["Plotting"]],
                t.li[t.a(href="/help")["Help"]]
                ])
            ]

        if not hasattr(self,"experiment") and not\
                hasattr(self,"experiment_id"):
            return ctx.tag[stan(
                t.ul[
                    t.li[t.a(href="/experiments")["Experiments"]],
                    t.li[t.a(href="/help")["Help"]]
                ])]
        else:
            return self.experiment.id.addCallback(respondToID)
