from fabric.api import *
from fabric import network, tasks
from fabric.colors import green, blue
import os
import sys
import logging
import socket
import uuid
import multiprocessing

#Make sure to add safe parent directory to $PATH
sys.path.append(os.path.join(os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "..")))

#Uses same settings as django
from django.conf import settings
os.environ.setdefault("DJANGO_SETTINGS_MODULE","web.config.settings")
 
#Import database models from django 
from web.safe.models import *

FORMAT = "%(name)s %(funcName)s:%(lineno)d %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO)

query = Worker.objects.filter(connected=0)
if query.count() > 0:
  env.hosts = query.values_list('host', flat=True)
else:
  env.hosts= []
  

#Get the Worker object that is currently being executed
def getWorkers():
  if env.host_string is not None:
    return Worker.objects.filter(host=env.host_string, connected = 0)
  else:
    print(blue('EXITED'))
    exit(1)


def workerFunction(worker, exp_description, installation, port):
  distribute(installation, worker)
  configure(installation, worker)
  startWorkers(port, exp_description, worker)
  
@task
@parallel
def runWorkers(experiment, port):
  from django.db import connection
  connection.close() 

  exp_description = experiment.exp_description
  installation = exp_description.installation

  workers = getWorkers()
  
  if len(workers) != 1:
    #close connecction so each child creates their own database connection
    connection.close() 
    q = multiprocessing.Queue()
    proc_list = []
    for worker in workers:
      proc_list.append(multiprocessing.Process(target=workerFunction, args=(worker, exp_description, installation, port)))
      proc_list[-1].start()    
    for proc in proc_list:
      proc.join()
  else:
    worker = workers[0]
    workerFunction(worker, exp_description, installation, port)
      
def distribute(installation, worker):
  """Distribute the compressed archive to all clients."""
  #If the installation was passed as an ID instead of an object,
  #retrieve object from database
  #This way is possible to use "fab distribute:INSTALLATION_ID"
  

  #Get current worker
  # worker=getWorker()
  worker.connected = 1
  worker.save(update_fields=['connected'])

  print(green("SAFE: Distributing '"+installation.name+"' to "+env.host))

  with cd(str(worker.path)):
    installation_id  = -1
    rebuild = False
    if os.path.isfile(worker.path+"/installation.txt"):
      with open(worker.path+"/installation.txt", 'r') as f:
        installation_id = int(f.read())
 
    if int(installation_id) != int(installation.id):
      print ("Id ",installation_id, "Doesn't match existing id",installation.id, "Rebuilding")
      print blue("Installation ID of worker does not match current installation. Rebuilding from tarball")
      rebuild = True
        
    #Append errors to this log file
    err_append = " 2>> error_log"
      
    #First, clean destination folder
    if rebuild:
      run('rm -rf *'+err_append)
      run('mkdir safe')
      with open(worker.path+"/installation.txt", 'w')  as f:
        f.write(str(installation.id))
    else:
      run('rm -rf safe'+err_append)
      run('mkdir safe')
        
    with cd("safe"):
      #Copy the necessary files to run the worker there
      run('touch __init__.py'+err_append)
      put(os.path.join(os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "worker")), ".")

    if rebuild:
      #Copy the tarball
      put(os.path.join(os.path.join(os.path.abspath(os.path.dirname(__file__)), "installations", str(installation.tarball_path))), ".")
      #Uncompress the tarball
      run('tar --strip=1 -zxf '+os.path.basename(str(installation.tarball_path))+err_append)
      #Remove tarball
      run('rm -f '+os.path.basename(str(installation.tarball_path))+err_append)

def configure(installation, worker):
  """ Prepare all clients. 
  
  Done by configuring the simulator install.
  
  """
  #If the installation was passed as an ID instead of an object,
  #retrieve object from database
  #This way is possible to use "fab distribute:INSTALLATION_ID"
  if not isinstance(installation, Installation):
    installation=Installation.objects.get(pk=installation)
    
  print(green("SAFE: Preparing "+env.host+" to execute '"+installation.name+"':"))
  with cd(str(worker.path)):
    run('./waf configure --enable-examples --enable-tests --disable-python > config_log 2>&1')

      
def startWorkers(port, exp_desc, worker):
  """ Start all workers."""
  
  print(green("SAFE: Starting all workers:"))
  #If the exp. desc. was passed as an ID instead of an object,
  #retrieve object from database
  #This way is possible to use "fab startWorkers:exp_desc_ID"
  if not isinstance(exp_desc, ExpDescription):
    exp_desc=ExpDescription.objects.get(pk=exp_desc)

  installation=exp_desc.installation

  server_hostname = socket.getfqdn()

  #Lets calculate the simulator path
  sim_path = os.path.abspath(os.path.join(worker.path, installation.simulator_folder))

  #Generate a key that will be used by the worker to authenticate later with us
  auth_key = str(uuid.uuid4())

  worker.auth_key = auth_key
  worker.connected = False
    
  worker.save(update_fields = ['auth_key','connected'])

  with cd(str(worker.path)):
    with cd("safe"):
      with cd("worker"):
        run("python2.7 worker.py "+server_hostname+" "+str(port)+" "+auth_key+" "+sim_path+" "+exp_desc.script+" "+str(worker.children))
        #run("python worker.py ")
  worker.connected = 0;
  worker.save(update_fields = ['connected'])
  print (green("SAFE: Worker executed"))
  import sys
  sys.stdout.flush()
