// The experiment currently being observed
var exp_id;
var colors = safe.colors.QUALITATIVE_0;
var focusDataPointRadius = 2;
var contextDataPointRadius = 1.5;
var numberFormat = d3.format(".2f");
var windowResizeFunctions = [];

var selectFactor;

//----------------------------------------------
// Plot Area Object
// ---------------------------------------------

MetricPlotArea = function(exp_id, factor, parentElement){
	this.exp_id = exp_id;
    this.factor = factor;
	this.parent = safe.tools.d3Object(parentElement);
    this.margin = {top: 10, right: 40, bottom: 50, left: 60};

	this.colorList = {};		// list of colors by run_id and metric_id
	this.displayList = {};		// list of metrics being displayed by run_id and metric_id
	
    this.graphWidth = undefined;
    this.contextGraphHeight = 180;
    this.focusGraphHeight = 500;
	this.plotList = {};
    this.designPointList = {};

    this.selectedGraph = undefined;
    this.contextGraph = undefined;

    d3.json("/safe/exp/" + exp_id + "/designpoints/", function(error, designPoints){
        if(error){ console.warn(error) }
        for(var index in designPoints){
            this.designPointList[designPoints[index].id] = designPoints[index];
        }
        this.initComponents();
    }.bind(this)); 
}

MetricPlotArea.prototype.initComponents = function(){
    this.dSelf = this.parent
		.append("div")
        .attr("id", this.id)
        .attr("class", "plot_area");

    var hider = new safe.tools.window_buttons(this.dSelf);
    hider.initComponents();

    windowResizeFunctions.push(this.resize.bind(this));

    this.main = this.dSelf.append("div").style({"width":"80%", "display":"inline-block"});
    var title = this.main.append("div").attr("class", "plot_area_label").text(this.factor.name);

    this.graphArea = this.main.append("div").attr({"height": this.contextGraphHeight + "px", "class": "graph_area"});
    this.graphWidth = this.graphArea[0][0].offsetWidth;

    this.select = this.dSelf.append("div").style({"width":"20%", "display":"inline-block", "vertical-align":"top"});
    this.select.append("div").attr("class", "plot_selection_label").text("Select");
    
	this.selection = new safe.selector.lineplot(this.exp_id, this.factor, this.selectionChange.bind(this), this.select);
    this.select.append("div").style("height","0.5cm");
    this.select.append("div").attr("class", "plot_selection_label").text("Options");

    this.options = new safe.selector.options(title, this.select);
    this.options.disableGraphMovement();

    this.displayed = this.main.append("div").attr("class", "plot_displayed");
    this.displayed.append("div").attr("class", "plot_displayed_label").text("Displayed Metrics");
    this.displayKey = new safe.infobox.colored(this.removeMetricFromDisplay.bind(this) , this.displayed);
 
    this.reserved = this.main.append("div").attr("class", "plot_reserved");
    this.reserved.append("div").attr("class", "plot_reserved_label").text("Reserved Colors");
    this.reservedColors = new safe.infobox.colored(this.removeReservedColor.bind(this), this.reserved);
   
    this.focus = new safe.graph.lineplot(this.margin, this.graphWidth, this.focusGraphHeight, undefined, this.graphArea, false);
    this.focus.setXAxisLabel(this.factor.name);
    this.options.setSelectedGraph(this.focus);

    this.context = new safe.graph.lineplot(this.margin, this.graphWidth, this.contextGraphHeight, undefined, this.graphArea, true);
    this.context.addBrush(this.brushed.bind(this), safe.constants.Y_AXIS);
    this.context.yAxis.ticks(5);
    this.options.setContextGraph(this.context);
    onWindowResize();
}

MetricPlotArea.prototype.getQueryParameters = function(levelDict){
    var parameters = [];
    Object.keys(levelDict).forEach(function(key){
        parameters.push({ key:levelDict[key] });
    });
    return parameters;
}

// levelDict -> a dictionary of (level_name, level) pairs
MetricPlotArea.prototype.selectionChange = function(levelDict, metric){
    var metric_id = metric.id;
    var keyString = getKeyString(levelDict);    

    if(this.currentMetric == undefined){ 
        this.currentMetric = metric;
        this.focus.setYAxisLabel(metric.name + "(" + metric.units + ")");
        this.options.setLabels();
    }

    if(metric.id != this.currentMetric.id){
        alert("Cannot display data from multiple metrics on the same plot. The current metric type is " + this.currentMetric.name + ".");
        return;
    }

    if(this.colorList[keyString] == undefined){ this.colorList[keyString] = {} }
	if(this.displayList[keyString] == undefined) { this.displayList[keyString] = {} }

	if(this.colorList[keyString][metric_id] == undefined) { 
        if(colors.length == 0){
            alert("All available colors have been used. Please remove one or more colors from the reserved colors list for further use.");
            return;
        }
        this.colorList[keyString][metric_id] = colors.shift() 
    }else{
        this.reservedColors.removeElement(keyString + "_" + metric_id);
    }	

	if(this.displayList[keyString][metric_id] == undefined){
		this.addDisplayMetric(levelDict, keyString, metric);
	}else{
		this.removeDisplayMetric(getDisplayString(levelDict, metric.name), keyString, metric);
	}
}

MetricPlotArea.prototype.addDisplayMetric = function(levelDict, keyString, metric){
    var metric_id = metric.id;
    var id = keyString + "_" + metric_id;

    var designPointIds = [];
    this.displayList[keyString][metric_id] = true;
    this.selection.addDisplayed(keyString, metric_id, this.colorList[keyString][metric_id]);

    // Updates the visual components for the graph being displayed
	var update = function(){
        this.context.displayAll();
        this.focus.displayAll();
        
        var displayString = getDisplayString(levelDict, metric.name);
        var color = this.colorList[keyString][metric_id];
        var returnAttributes = {"keyString": keyString, "metric": metric, "displayString": displayString };

        this.displayKey.addElement(id, displayString, color, returnAttributes)
        this.displayKey.updateElements();
        this.context.brushed();
    }.bind(this)

    // Creates a new graph from statisticalData
    var createDisplay = function( data ){

        pointAttributes = {"stroke": this.colorList[keyString][metric_id], 
                           "stroke-width": focusDataPointRadius, 
                           "fill": "white", 
                           "fill-opacity": 0, 
                           "r": focusDataPointRadius};
        
        this.focus.addDataSet(id, data, function(d){ return d.time }, function(d){ return d.value }, pointAttributes);
        this.context.addDataSet(id, data, function(d){ return d.time }, function(d){ return d.value }, pointAttributes);

        update();
    }.bind(this)

    // Determine whether a new graph must be created
    if(!this.focus.hasDataSet(id)){
        var url = "/safe/exp/" + this.exp_id + "/factor/" + this.factor.name + "/metric/" + metric_id + "/results/";
        d3.json(url).post(JSON.stringify({factors:levelDict}), function(error, data){
            if(error){ console.warn(error) }

            var statisticalData = calculateStats(data);
            createDisplay(statisticalData);
        });
    }else{
        this.focus.setVisible(id);
        this.context.setVisible(id);
		update();
    }

}

MetricPlotArea.prototype.removeMetricFromDisplay = function(attributes){
    this.removeDisplayMetric(attributes.displayString, attributes.keyString, attributes.metric);
}

MetricPlotArea.prototype.removeDisplayMetric = function(displayString, keyString, metric){
    var metric_id = metric.id;
    var id = keyString + "_" + metric_id;
    var returnAttributes = {"displayString": displayString, "keyString":keyString, "metric":metric};

    this.displayKey.removeElement(id);
    this.reservedColors.addElement(id, displayString, this.colorList[keyString][metric_id], returnAttributes);
    this.selection.removeDisplayed(keyString, metric_id);

    this.focus.setInvisible(id);
    this.context.setInvisible(id);
    this.context.brushed();

    delete this.displayList[keyString][metric_id];

    if(this.focus.getVisibleIds().length == 0){ this.currentMetric = undefined }
}

MetricPlotArea.prototype.removeReservedColor = function(attributes){
    var keyString = attributes.keyString;
    var metric_id = attributes.metric.id;
    var id = keyString + "_" + metric_id;


    colors.push(this.colorList[keyString][metric_id]);
    delete this.colorList[keyString][metric_id];
    this.reservedColors.removeElement(id);

    this.focus.remove(id);
    this.context.remove(id);
}

MetricPlotArea.prototype.onGraphSelect = function(graph, src){
    this.selectedGraph = graph;
    this.context.removeAll();

    var id = graph.getVisibleIds()[0];
    var dataSet = graph.getDataSet(id);
    var pointAttributes = graph.getPointAttributes(id);
    
    this.context.addDataSet(id, dataSet, pointAttributes);
    this.context.display(id);

    this.options.setSelectedGraph(graph);
    if(src == undefined){ src = d3.event.target }
    var cxt = this.context.svg[0][0];
    cxt.parentNode.insertBefore(cxt, src.nextSibling);
}

MetricPlotArea.prototype.brushed = function(brush) {
    var extent = brush.extent();
    xDomain = brush.empty() ? this.context.xScale.domain() : [ extent[0], extent [1] ];

    this.focus.setXDomain(xDomain);
    this.focus.displayAll();
}

MetricPlotArea.prototype.resize = function(){
    var graphArea = this.graphArea[0][0];
    if(graphArea == undefined){ return }

    this.graphWidth = graphArea.offsetWidth;
    this.context.setWidth(this.graphWidth);
    this.context.displayAll();

    this.focus.setWidth(this.graphWidth);
    this.focus.displayAll();
}

//======================================================

function onWindowResize(){
    windowResizeFunctions.forEach(function(resizeFunction){
        resizeFunction();
    });
}

function initVisualComponents(){
    d3.select(window).on("resize", onWindowResize);
	selectFactor = new safe.selector.factor_chooser(exp_id, addFactorPlotArea, "select_factor");
}

function addFactorPlotArea(factor){
	new MetricPlotArea(exp_id, factor, "place_graphs");
    d3.select("#place_graphs").append("div").attr("class", "graph_sep");
}

function getKeyString(levelDict){
    var keyString = "";
    Object.keys(levelDict).sort().forEach(function(key){
        keyString += key + "_" + levelDict[key] + "_";
    });
    keyString = keyString.split(".").join("_");
    return keyString;
}

function getDisplayString(levelDict, metric_name){
    var displayString = metric_name;
    Object.keys(levelDict).sort().forEach(function(key){
        displayString += ", " + key + " = " + levelDict[key]
    });
    return displayString;
}

function calculateStats(data){
    var levelData = {};

    data.forEach(function(d){
        if(levelData[d.level] == undefined){ levelData[d.level] = [] }
        levelData[d.level].push(d);
    });
            
    var stats = [];
    d3.entries(levelData).forEach(function(d){
        var currentStats = {};
        var confidence = safe.tools.confidence(d.value, 1.645, function(d) { return d.value });

        currentStats["x"] = d.key;
        currentStats["lowY"] = confidence[0];
        currentStats["midY"] = confidence[1];
        currentStats["highY"] = confidence[2];

        stats.push(currentStats);
    });
        
    return stats;
}

function readParameters(){
    var parameterList = location.search.substring(1).split("&");
    var parameters = {};
    parameterList.forEach(function(parameter){
        var temp = parameter.split("=");
        parameters[temp[0]] = temp[1];
        });

    exp_id = parameters["exp_id"] == undefined ? 1: parameters["exp_id"] ;
}

readParameters();
initVisualComponents();
