/**--------------------------------------
 * A graph for time series
 * -------------------------------------
 *
 * This graph take data with {time:__, value:__} pairs and
 * graphs them as a time series
 */

safe.graph.timeplot = safe_graph_timeplot;

// This plot inherits from the cartesian plot
safe_graph_timeplot.prototype = Object.create(safe.graph.cartesian.prototype);
safe_graph_timeplot.prototype.constructor = safe_graph_timeplot;

/**
 * Creates a new lineplot object
 * @param margins The padding to be placed around the graphing area {top: __, bottom:__, right:__, left:__}
 * @param width The pixel width of the entire graph including graphing area and margins
 * @param selectedCallBack The function to be called when the user clicks on this graph
 *          This value can be "undefined" in which case nothing will happen (will be called "selectedCallBack(this)")
 * @param parentElement The element to which this graph should be added. This can be
 *          a dom element, an html id, or a d3 object.
 * @param displayPoints If false or undefined, points are not displayed on lines. Otherwise, points are displayed
 *
 * @return The newly created object
 * */
function safe_graph_timeplot(margins, width, height, selectedCallBack, parentDomId, displayPoints){
    safe.graph.cartesian.apply(this, [margins, width, height, selectedCallBack, parentDomId]);

    this.displayPoints = displayPoints == undefined || displayPoints != false ? true : displayPoints;

    this.line = d3.svg.line()
        .x(function(d) { return this.margin.left + this.xScale(this.xAccessor[id](d)) }.bind(this) )
        .y(function(d) { return this.margin.top + this.yScale(this.yAccessor[id](d)) }.bind(this) );

    this.numberFormat = d3.format(".2f");
}

/**
 * Displays the specified data set in the form associated with line graphs
 * @param id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_timeplot.prototype.display = function(id){
    if(this.toDisplay[id] != true) { 
        this.svg.selectAll(".circle_" + id).remove();
        this.svg.selectAll(".path_" + id).remove();
        return;
    }

    var paths = this.svg.selectAll(".path_" + id).data([this.data[id]]);

    paths.enter().append("path")

    paths
        .attr("d", function(d){ return this.line(d) }.bind(this))
        .attr("clip-path", "url(#" + this.clipId + ")" )
        .attr("class", "path_" + id)
        .attr({"stroke-width":"1", "stroke":this.pointAttributes[id].stroke });

    paths.exit().remove();

    if(this.displayPoints == true){
        var circles = this.svg.selectAll(".circle_" + id).data(this.data[id]);

        circles.enter().append("circle")
            .attr("class", "circle_" + id)
            .attr(this.pointAttributes[id])
            .attr("clip-path", "url(#" + this.clipId + ")" );
        
        var counter = 0;
        circles.each(function(d) { if(this.xAccessor[id](d) > this.xScale.domain()[0] && this.xAccessor[id](d) < this.xScale.domain()[1]) counter++; }.bind(this));
        var scaleFactor = this.getScaleFactor(0.1, counter);
        var position = 0;

        circles 
            .attr("cx", function(d) { return this.margin.left + this.xScale(this.xAccessor[id](d)) }.bind(this))
            .attr("cy", function(d) { return this.margin.top + this.yScale(this.yAccessor[id](d)) }.bind(this))
            .attr("info", function(d) { return this.numberFormat(this.xAccessor[id](d)) + "<br/>" + this.numberFormat(this.yAccessor[id](d)) }.bind(this))
            .attr("visibility", function(d, i){
                var t;
                if((i % scaleFactor == 0) && this.xScale(this.xAccessor[id](d)) >= position) {
                    t = "visible";
                    position = this.xScale(this.xAccessor[id](d)) + 5;
                } else {
                    t = "hidden";
                }
                return t;
                }.bind(this))
            .on("mouseover", this.mouseOverPoint.bind(this))
            .on("mouseout", this.mouseOutPoint.bind(this));

       circles.exit().remove();
    }
}

/** 
 * Adds the specified data set to this graph (must be followed by display() or displayAll())
 * @param id A string that distinguishes this data set from all others on this graph
 * @param data The data to be added to this graph
 * @param pointAttributes A dictionary of html attributes to add to each displayed point
 * @param mouseEvents A dictionary of mouse events to occur when mouse is over a data point
 */
safe_graph_timeplot.prototype.addDataSet = function(id, data, pointAttributes, mouseEvents){
    data.sort(function(a, b){ return a.time - b.time });
    this.data[id] = data;
    this.pointAttributes[id] = pointAttributes;
    this.toDisplay[id] = true;
    this.updateDataFeatures();
}

/**
 * Gets the scale factor for points on the graph. This number essentially means
 * plot every x-th point.
 *
 * @param pointsPerPixel How many points to plot per pixel (recommended 10)
 * @param size The number of points being plotted
 * @return the scale factor as an integer
 */
safe_graph_timeplot.prototype.getScaleFactor = function(pointsPerPixel, size) {
    return Math.ceil((size / this.width) * pointsPerPixel);
}
