/**----------------------------------
 * A graph for histograms
 *----------------------------------
 */

safe.graph.histogram = safe_graph_histogram;

safe_graph_histogram.prototype = Object.create(safe.graph.base.prototype);
safe_graph_histogram.constructor = safe_graph_histogram;

/**
 * Creates a new histogram object
 * @param margins The padding to be placed around the graphing area {top: __, bottom:__, right:__, left:__}
 * @param width The pixel width of the entire graph including graphing area and margins
 * @param selectedCallBack The function to be called when the user clicks on this graph
 *          This value can be "undefined" in which case nothing will happen (will be called "selectedCallBack(this)")
 * @param parentElement The element to which this graph should be added. This can be
 *          a dom element, an html id, or a d3 object.
 *
 * @return The newly created object
 * */
function safe_graph_histogram(margins, width, height, onSelectedCallBack, parentElement){
    safe.graph.base.apply(this, [margins, width, height, onSelectedCallBack, parentElement]);

    this.frequency = false;
    this.infoFormat = d3.format(".4f");
    this.tickFormat = d3.format(".2f");
    this.bins = 5;
    this.tooltip = d3.select("body").append("div")
        .attr("class", "tooltip")
        .style("opacity", 0);

    return this;
}

/** Redisplays all data sets for this graph */
safe_graph_histogram.prototype.displayAll = function(){
    this.display();
}

/**
 * Displays the specified data set in the form associated with line graphs
 * @param id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_histogram.prototype.display = function(){
    if(this.histogram == undefined){ return }

    var bars = this.svg.selectAll(".bar").data(this.histogram);
    bars.enter().append("g").attr("class", "bar");
    bars.exit().remove();

    bars.selectAll("rect").remove();
    bars.append("rect");

    var rects = bars.selectAll("rect");
    rects
        .attr("x", function(d) { return this.margin.left + this.xScale(d.x  + 0.05 * d.dx) }.bind(this))
        .attr("y", function(d) { return this.margin.top + this.yScale(d.y)}.bind(this))
        .attr("width", function(d) { return this.xScale(this.xScale.domain()[0] + d.dx) * 0.9 }.bind(this))
        .attr("height", function(d) { return this.height - this.yScale(d.y) }.bind(this))
        .attr("info", function(d) { return this.infoFormat(d.y) + "</br>" + 
                                    this.infoFormat(d.x) + " - " + this.infoFormat(parseFloat(d.x) + parseFloat(d.dx)) }.bind(this))
        .style(this.style)
        .on("mouseover", this.mouseOverBox.bind(this))
        .on("mouseout", this.mouseOutBox.bind(this));

    bars.selectAll("text").remove();
    bars.append("text");

    var text = bars.selectAll("text");
    text
        .attr("x", function(d) { return this.margin.left + this.xScale(d.x) }.bind(this))
        .attr("y", function(d) { return this.margin.top + this.yScale(d.y)}.bind(this))
        .attr("dy", "20px")
        .attr("dx", function(d) { return this.xScale(this.xScale.domain()[0] + (d.dx/2)); }.bind(this))
        .attr("fill", "#fff")
        .attr("text-anchor", "middle")
        .attr("pointer-events", "none")
        .text(function(d) { return this.tickFormat(d.y) }.bind(this));

}

/**
 * Sets the data to be dispayed on this graph
 * @param data A list of values to be used in the creation of the graph
 * @param style A dictionary of css attributes that will be applied
 *        to each of the rectangles (bars) of the graph
 */
safe_graph_histogram.prototype.setData = function(data, style){
    if(data == undefined){ return }

    this.data = data;
    this.style = style;
    this.extent = d3.extent(data)
    this.setXDomain(this.extent);
    this.update();
}

/** Updates the content of this graph based on current instance values */
safe_graph_histogram.prototype.update = function(){
    this.histogram = d3.layout.histogram().frequency(this.frequency).bins(this.bins).range(this.xScale.domain())(this.data);
    this.setYDomain([0, d3.max(this.histogram, function(d) { return d.y })]);

    var tempScale = d3.scale.linear().domain([0, this.bins]).range(this.xScale.domain());
    var tickArray = d3.range(this.bins + 1).map(tempScale);
    this.xAxis.tickValues(tickArray).tickFormat(this.tickFormat);
    this.svg.select(".x.axis").call(this.xAxis);
}

/** Returns the data currently contained in the graph */
safe_graph_histogram.prototype.getData = function(){
    return this.data;
}

/** Returns the range of values currently associated with this graph */
safe_graph_histogram.prototype.getExtent = function(){
    return this.extent;
}

/** Removes the currently displayed data from the graph */
safe_graph_histogram.prototype.remove = function(){
    this.svg.selectAll(".bar").remove();
}

/** 
 * Sets the number of bins to the nearest number of bins by which the 
 * bins will be of equal size
 *
 * Note: The number of bins used may not be exactly what is specified
 *
 * @param bins The number of bins desired
 */
safe_graph_histogram.prototype.setBinNumber = function(bins){
    this.bins = bins;
    this.update();
}

/**
 * Returns the actual number of bins currently being used
 * @return The number of bins currently being used
 */
safe_graph_histogram.prototype.getBinNumber = function(){
    return this.bins;
}

/**
 * Sets the approximate width of the bins. The width that will
 * be used is the closest width that will provide equal width
 * to all bins
 *
 * @param width The desired width of the bins
 */
safe_graph_histogram.prototype.setBinWidth = function(width){
    this.bins = Math.round(d3.max(this.data) / width);
    if(this.bins <= 0){ this.bins = 1 }
    this.update();
}

/**
 * Returns the actual width of the bins currently being used
 * @return The width of the bins currently bing used
 */
safe_graph_histogram.prototype.getBinWidth = function(){
    if(!(this.histogram && this.histogram.length > 0)){ return 0 }
    return this.histogram[0].dx;
}

/**
 * Sets whether the values are calculated by as frequency or percentage
 * @param frequency A boolean which is true if the desired measurement
 *        is frequency
 */
safe_graph_histogram.prototype.setFrequency = function(frequency){
    this.frequency = frequency;
    this.update();
}

/**
 * Returns the boolean representing whether frequency or percentage is being used
 * @return true if frequency is being used, false if percentage is being used
 */
safe_graph_histogram.prototype.getFrequency = function(){
    return this.frequency;
}

/** Called when mouse is over a histogram box */
safe_graph_histogram.prototype.mouseOverBox = function(){
    this.tooltip.transition()
        .duration(200)
        .style("opacity", 0.9);

    this.tooltip.html(d3.event.target.getAttribute("info"))
        .style("left", (d3.event.pageX) + "px")
        .style("top", (d3.event.pageY - 28) + "px");
}

/** Called when mouse is no longer over a histogram box */
safe_graph_histogram.prototype.mouseOutBox = function(){
    this.tooltip.transition()
        .duration(200)
        .style("opacity", 0);
}
