/**-------------------------------------
 * A full selector object
 * -------------------------------------
 *
 * An object prototype which includes all selection capabilities
 * currently in place.
 */
safe.selector.full = safe_selector_full;

/**
 * Creates a new full selector object
 * @param exp_id The experiment id with which this selector is to be associated
 * @param selectionCallBack A function to be called when a metric is selected
 * @param applyFilterCallBack A function to be called when a new filter is created
 * @param parementElement The dom id, dom element, or d3 object to which this selector
 *        should be added
 */
function safe_selector_full(exp_id, selectionCallBack, applyFilterCallBack, parentElement, callBack){
        this.exp_id = exp_id;
        this.dParent = safe.tools.d3Object(parentElement);
        this.selectionCallBack = selectionCallBack;
        this.applyFilterCallBack = applyFilterCallBack;
        
        this.dSelf = this.dParent.append("div")
            .attr("class", "well container")
            .style("width", "95%");

        this.levelRangeLeft = document.createElement("div");
        this.levelRangeRight = document.createElement("div");

        this.hider = new safe.tools.hider(this.dSelf, this.dParent);
        this.selectFactor = new safe.selector.factor(this.exp_id, this.factorSelectChange.bind(this), this.dSelf);
        this.selectLevelMin = new safe.selector.level(this.exp_id, undefined, this.levelRangeLeft);
        this.selectLevelMax = new safe.selector.level(this.exp_id, undefined, this.levelRangeRight);
        this.selectDesignPoint = new safe.selector.designpoint(this.exp_id, this.designPointSelectChange.bind(this), this.dSelf);
        this.selectRun = new safe.selector.run(this.exp_id, this.runSelectChange.bind(this), this.dSelf);
        this.selectMetric = new safe.selector.metric(this.exp_id, this.selectElementChange.bind(this), this.dSelf);
        this.selectResults = new safe.selector.results(this.dSelf);

        this.selectFactor.readInitialData(function(){
            this.selectLevelMin.readInitialData(function(){
                this.selectLevelMax.readInitialData(function(){
                    this.selectMetric.readInitialData(function(){
                        this.initComponents();
                        if(callBack){ callBack() }
                    }.bind(this));
                }.bind(this));
            }.bind(this));
        }.bind(this));
}
      
/**
 * Initializes the dom components for this object
 */
safe_selector_full.prototype.initComponents = function(){
    this.hider.initComponents();

    // Initialize the factor list
    this.selectFactor.initComponents();
    this.selectLevelMin.initComponents();
    this.selectLevelMax.initComponents();

    this.dSelf[0][0].appendChild(this.levelRangeLeft);
    this.dSelf[0][0].appendChild(this.levelRangeRight);
    d3.select(this.levelRangeLeft).attr("class", "col-xs-6").style("padding", "0px");
    d3.select(this.levelRangeRight).attr("class", "col-xs-6").style("padding", "0px");
     
    // Initialize the apply selection button
    this.dSelf
        .append("button")
        .attr("class", "btn btn-default col-xs-12")
        .text("Apply Selection Filter")
        .on("click", this.applyDesignFilter.bind(this));

    this.selectDesignPoint.initComponents();
    this.selectRun.initComponents();
    this.selectMetric.initComponents();
    this.selectResults.initComponents();

    this.factorSelectChange();
    this.updateDesignPoints();
}

/**
 * Called when the selected factor is changed
 */
safe_selector_full.prototype.factorSelectChange = function(){

    var factor_id = this.selectFactor.getSelectedFactor().id;

    this.selectLevelMin.updateLevels(factor_id);
    this.selectLevelMax.updateLevels(factor_id);

    this.selectLevelMax.eSelect.selectedIndex = this.selectLevelMax.eSelect.options.length - 1;
}

/**
 * Called when the "Apply Filter" button is clicked
 */
safe_selector_full.prototype.applyDesignFilter = function(){

    var factor = this.selectFactor.getSelectedFactor();
    var min = this.selectLevelMin.getMinLevel(factor.id);
    var max = this.selectLevelMax.getMaxLevel(factor.id);

    if(this.getMinLevel() > this.getMaxLevel() ){
        alert("The factor's minimum level cannot be greater than the factor's maximum level.");
        return;
    }

    var parameter = {"name":factor.name, "factor_id":factor.id, "min_value":this.getMinLevel(), "max_value":this.getMaxLevel() };
    if(this.getMinLevel() > min || this.getMaxLevel() < max ){
        this.selectDesignPoint.addParameter(factor.id, parameter);
    }else{
        this.removeDesignFilter(factor.id);
    }

    this.updateDesignPoints();
    this.applyFilterCallBack(parameter);
}

/**
 * Removes the specified filter from the list of filters
 * @param factor_id The id of the factor with which the filter is associated
 */
safe_selector_full.prototype.removeDesignFilter = function(factor_id){
    this.selectDesignPoint.removeParameter(factor_id);

    this.updateDesignPoints();
}

/**
 * Updates the listed design points based on filters and information in the database
 */
safe_selector_full.prototype.updateDesignPoints = function(){
    this.selectDesignPoint.updateDesignPoints(this.designPointSelectChange.bind(this));
}

/**
 * Called when the selected design point changes
 */
safe_selector_full.prototype.designPointSelectChange = function(){
    this.selectRun.updateRuns(this.selectDesignPoint.getSelectedDesignPoint(), function(){
        this.selectMetric.setDistinctionId(this.getSelectedRun());
        this.selectMetric.updateSelectedMetrics();  
    }.bind(this));
}

/**
 * Called when the selected run is changed
 */
safe_selector_full.prototype.runSelectChange = function(){
    this.selectMetric.setDistinctionId(this.getSelectedRun());
    this.selectMetric.updateSelectedMetrics();
}

/**
 * Adds a metric to the list of displayed metrics and associates a color with it
 * @param run_id The id of the run with which the metric is associated
 * @param metric_id The id of the metric being displayed
 * @param color The color to be associated with the metric
 */
safe_selector_full.prototype.addDisplayed = function(run_id, metric_id, color){
    this.selectMetric.setMetricColor(run_id, metric_id, color);
    this.selectMetric.updateSelectedMetrics();
}

/**
 * Returns the parameters being used to determine selectable design points
 * @return The parameters being used to determine selectable design points, form:
 *         [{"factor_name": <factor_name>, "min_value":<min_value>, "max_value":<max_value>}, ...]
 */
safe_selector_full.prototype.getDesignParameters = function(){
    return this.selectDesignPoint.getSelectionParameters();
}

/**
 * Removes a metric from the list of displayed metrics
 * @param run_id The id of the run with which the metric is associated
 * @param metric_id The id of the metric to be removed
 */
safe_selector_full.prototype.removeDisplayed = function(run_id, metric_id){
    this.selectMetric.setMetricColor(run_id, metric_id, undefined);
    this.selectMetric.updateSelectedMetrics();
}

/**
 * Called when a metric is selected from the metric selector component
 */
safe_selector_full.prototype.selectElementChange = function(){
    var metric_id = d3.event.target.getAttribute("metric_id");
    var metric = this.selectMetric.getMetric(metric_id);
    if(this.selectionCallBack){ this.selectionCallBack(this.getSelectedRun(), metric) }
}

/**
 * Returns the currently selected factor
 * @return The curretnly selected factor
 */
safe_selector_full.prototype.getSelectedFactor = function(){
    return this.selectFactor.getSelectedFactor();
}

/**
 * Returns the currently selected minimum level
 * @return The curretnly selected minimum level
 */
safe_selector_full.prototype.getMinLevel = function(){
    return this.selectLevelMin.getSelectedLevel();
}

/**
 * Returns the currently selected maximum level
 * @return The currently selected maximum level
 */
safe_selector_full.prototype.getMaxLevel = function(){
    return this.selectLevelMax.getSelectedLevel();
}

/**
 * Returns the currently selected design point
 * @return the currently selected design point
 */
safe_selector_full.prototype.getSelectedDesignPoint = function(){
    return this.selectDesignPoint.getSelectedDesignPoint();
}

/**
 * Returns the currently selected run
 * @return the currently selected run
 */
safe_selector_full.prototype.getSelectedRun = function(){
    return this.selectRun.getSelectedRun();
}

/**
 * Returns the parameters currently being used to filter results
 * @return filter parameters of the form [{<parameter_name>:<parameter_value>}, ...]
 */
safe_selector_full.prototype.getResultsParameters = function(){
    return this.selectResults.getResultsParameters();
}
