/**-------------------------------------
 * A run selector object 
 * -------------------------------------
 *
 * An object prototype which allows the selection of runs 
 * which are stored in the database.
 */
safe.selector.run = safe_selector_run;

/**
 * Creates a new selection object
 * @param exp_id The experiment id with which this selector is associated
 * @param changeCallBack A function to be called when the selection is changed
 * @param parentElement The dom element, dom id, or d3 object to which this
 *        selector should be added.
 */
function safe_selector_run(exp_id, changeCallBack, parentElement){
    this.exp_id = exp_id;
    this.callBack = changeCallBack;
    this.dParent = safe.tools.d3Object(parentElement);
    this.runsByDesignPoint = {};
}

/**
 * Reads data from the database about available runs 
 * @param desinpointId The design point for which the runs are to be read
 * @param callBack A function to be called when the read is complete
 */
safe_selector_run.prototype.readData = function(designpointId, callBack){
    d3.json("/safe/designpoint/" + designpointId + "/simruns/", function(error, data){
        if(error) { console.warn(error) }

        this.runsByDesignPoint[designpointId] = data;
        callBack();
    }.bind(this));
}

/**
 * Initializes the visual dom elements for this selector
 */
safe_selector_run.prototype.initComponents = function(attributes){
    this.dSelect = this.dParent
        .append("select")
        .attr(attributes)
        .attr("class", "form-control")
        .on("change", this.selectChange.bind(this));
    this.eSelect = this.dSelect[0][0];
}

/**
 * Updates the currently displayed runs based on data in the database
 * @param desinpointId The design point for which the runs are to be displayed 
 * @param callBack A function to be called after the update is complete
 */
safe_selector_run.prototype.updateRuns = function(designpointId, callBack){
    var update = function(){
        runs = this.dSelect.selectAll("option").data(this.runsByDesignPoint[designpointId]);
        runs.enter().append("option");
        runs.attr("run", function(d) { return d.simrun_id }).text(function(d) { return "Run " + d.simrun_id });
        runs.exit().remove();
        callBack();
    }.bind(this);
    
    if(this.runsByDesignPoint[designpointId] == undefined){
        this.readData(designpointId, update);
    }else{
        update();
    }
}

/**
 * Called when there is a change in the currently selected run
 */
safe_selector_run.prototype.selectChange = function(){
    this.callBack(this.getSelectedRun());
}

/**
 * Returns the currently selected run
 * @return The curretnly selected run
 */
safe_selector_run.prototype.getSelectedRun = function(){
    return this.eSelect.options[this.eSelect.selectedIndex].getAttribute("run");
}
