/**-------------------------------------
 * Various SAFE tools 
 * -------------------------------------
 *
 * A collection of helpful functions for the SAFE
 * visualization tools
 */

safe.tools.confidence = safe_tools_confidence;
safe.tools.confidenceError = safe_tools_confidenceError;
safe.tools.css2json = safe_tools_css2json;
safe.tools.d3Object = safe_tools_d3Object;
safe.tools.embedCSS = safe_tools_embedCSS;
safe.tools.downloadElement = safe_tools_downloadElement;
safe.tools.getCSS = safe_tools_getCSS;
safe.tools.qqplot = safe_tools_qqplot;
safe.tools.qqnorm = safe_tools_qqnorm;
safe.tools.quartiles = safe_tools_quartiles;
safe.tools.stdev = safe_tools_stdev;
safe.tools.variance = safe_tools_variance;


/**
 * Selects a d3 object based on the given element
 * @param element A dom ID, d3 object, or dom element to be found
 * @return A d3 object for the specified element
 */
function safe_tools_d3Object(element){
    if(element.tagName){
        return d3.select(element);
    }else if(typeof(element) == "string" || element instanceof String){
        return d3.select("#" + element);
    }else{
        return element;
    }
}

/**
 * Returns the variance of a set of numbers
 * @param data A list of objects or numbers
 * @param accessor A function that, when called on on of the objects,
 *          will return a numerical value to be used in the computation.
 *          If this value is undefined, it is assumed that the objects
 *          themselves are numerical.
 */
function safe_tools_variance(data, accessor){
    if(!accessor){ accessor = function(d){ return d } }

    var mean = d3.mean(data, accessor);
    return d3.sum(data, function(d) { return Math.pow(accessor(d) - mean, 2 ) / (data.length - 1) });
}

/**
 * Returns the standard deviation of a set of numbers
 * @param data A list of objects or numbers
 * @param accessor A function that, when called on on of the objects,
 *          will return a numerical value to be used in the computation.
 *          If this value is undefined, it is assumed that the objects
 *          themselves are numerical.
 */
function safe_tools_stdev(data, accessor){
    return Math.sqrt(safe.tools.variance(data, accessor));
}

/**
 * Returns the confidence error
 * @param zScore The zScore for the data set
 * @param dev The deviation for the set of numbers
 * @param n The number of elements in the set
 */
function safe_tools_confidenceError(zScore, dev, n){
    return zScore * dev / Math.sqrt(n);
}

/**
 * Calculates the confidence intervals for the given data.
 * @paran data The data for which the intervals are to be calculated
 * @param zScore The zScore representing the desired confidence
 * @param xAccessor A function used on each element to extract the x value
 * @param yAccessor A function used on each element to extract the y value
 * @return An array of elements by x-value of the form [{"x":__, "lowY":__, "midY":__, "highY":__}, ...]
 */
function safe_tools_confidence(data, zScore, accessor){
    if(!accessor){ accessor = function(d){ return d } }
    
    returnData = [];
    var mean = d3.mean(data, accessor);
    var stdev = safe.tools.stdev(data, accessor);
    var error = safe.tools.confidenceError(zScore, stdev, data.length);

    return [ mean - error, mean, mean + error ];
}

/**
 * Calculates the quartiles for the given data.
 * @paran data The data for which the intervals are to be calculated
 * @param xAccessor A function used on each element to extract the x value
 * @param yAccessor A function used on each element to extract the y value
 * @return An array of elements by x-value of the form [{"x":__, "minY":__, "lowY":__, "midY":__, "highY":__, "maxY":__}, ...]
 */
function safe_tools_quartiles(data, accessor){
    if(!accessor){ accessor = function(d){ return d } }

    currentData = data.map(function(d){ return accessor(d) })
    currentData.sort(function(a, b){ return a - b });
    returnData = [];

    returnData.push(d3.quantile(currentData, 0));
    returnData.push(d3.quantile(currentData, 0.25));
    returnData.push(d3.quantile(currentData, 0.5));
    returnData.push(d3.quantile(currentData, 0.75));
    returnData.push(d3.quantile(currentData, 1));

    return returnData;
}

/**
 * Computes the points of a qqplot for a given set of data
 * with a normal distribution
 * @param givenData A list of objects or numbers
 * @param accessor A function that, when called on on of the objects,
 *          will return a numerical value to be used in the computation.
 *          If this value is undefined, it is assumed that the objects
 *          themselves are numerical.
 * @return A list of values in the form [{"x":__, "y":__, "norm":__}, ...]
 */
function safe_tools_qqnorm(givenData, accessor){
    if(!accessor){ accessor = function(d){ return d } }

    var results = [];
    var hypothetical = [];
    var data = [];

    givenData.forEach(function(d){ data.push(accessor(d))});
    data.sort(function(a, b){ return a - b });
    
    var mean = d3.mean(data);
    var n = data.length;
    var stdev = safe.tools.stdev(data);
    //var myString = "";
    //data.forEach(function(d){ myString += d + "\n" });
    //console.log(myString);

    data.forEach(function(value, index){
        var prob = (index + 1) / (n + 1);
        var normQuantile = normsInv(prob, mean, stdev);
        var devs = (normQuantile - mean)/stdev;
        results.push({"x":devs, "y":value, "norm":normQuantile});
    });
    return results;
}

/**
 * Computes the points of a qqplot for a one data set against the other 
 *
 * @param dataSet1 A list of objects or numbers
 * @param dataSet2 A list of objects or numbers
 * @param accessor A function that, when called on on of the objects,
 *          will return a numerical value to be used in the computation.
 *          If this value is undefined, it is assumed that the objects
 *          themselves are numerical.
 * @return A list of values in the form [{"x":__, "y":__}, ...]
 */
function safe_tools_qqplot(dataSet1, dataSet2, accessor){
    if(!accessor){ accessor = function(d){ return d } }

    var results = [];
    var data1 = [];
    var data2 = [];

    dataSet1.forEach(function(d){ data1.push(accessor(d))});
    dataSet2.forEach(function(d){ data2.push(accessor(d))});

    data1.sort(function(a, b){ return a - b });
    data2.sort(function(a, b){ return a - b });

    var n = data1.length;
    data1.forEach(function(value, index){
        var prob = (index + 1) / (n + 1);
        var quan1 = d3.quantile(data1, prob);
        var quan2 = d3.quantile(data2, prob);
        results.push({"x":quan1, "y":quan2});
    });

    return results;
}

/**
 * Takes the given element and includes all associated css styles
 * into the html code. The result is then given for download.
 * @param element The dom element to be downloaded (or jQuery selector for the element)
 * @param fileName The name to be given to the downloaded file
 */
function safe_tools_downloadElement(element, fileName){
    var serializer = new XMLSerializer;
    var element = element.cloneNode(true);
    safe.tools.embedCSS($(element));
    $(element).css("border", "1px solid black");

    var data = "text/svg;charset=utf-8," + serializer.serializeToString(element);
    var link = document.createElement("a");
    link.setAttribute("href", "data:" + data);
    link.setAttribute("download", fileName);
    link.click();
}

/**
 * Takes dom element and embeds all associated css styles into
 * the html code. This allows an element to be downloaded without
 * the loss of style components.
 * @param element The element to embed or its jquery selector
 */
function safe_tools_embedCSS(element){
    var style = safe.tools.getCSS($(element));
    $(element).css(style);
    var children = element.children("*");
    if(children.length > 0){
        children.each(function(i, d){ safe.tools.embedCSS($(d)) });
    }
}

/**
 * Determines all css styles associated with a given element
 * @param a A jquery object for which the css is to be found
 * @return A javascript object containing the styles of the object
 */
function safe_tools_getCSS(a){
    var sheets = document.styleSheets, o = {};
    for (var i = 0; i < sheets.length; i++){
        if(sheets[i].href.substring(sheets[i].href.length - 13) == "bootstrap.css"){
            continue;
        }
        var rules = sheets[i].rules || sheets[i].cssRules;
        for(var r = 0; r < rules.length; r++){
            if( a.is(rules[r].selectorText)){
                o = $.extend(o, 
                             safe.tools.css2json(rules[r].style),
                             safe.tools.css2json(a.attr("style")));
            }
        }
    }
    return o;
}

/**
 * Converts a css style to a javascript/json object
 * @param style The style to be converted
 * @return The associated javascript object
 */
function safe_tools_css2json(style){
    var s = {};
    if(!style) return s;
    if(style instanceof CSSStyleDeclaration){
        for(var i in style){
            if((style[i]).toLowerCase){
                s[(style[i]).toLowerCase()] = (style[style[i]]);
            }
        }
    } else if (typeof style == "string"){
        style = style.split("; ");
        for(var i in style){
            var x = style[i].split(": ");
            s[x[0].toLowerCase()] = (x[1]);
        }
    }
    return s;
}

/// Original C++ implementation found at http://www.wilmott.com/messageview.cfm?catid=10&threadid=38771
/// C# implementation found at http://weblogs.asp.net/esanchez/archive/2010/07/29/a-quick-and-dirty-implementation-of-excel-norminv-function-in-c.aspx
/*
 *     Compute the quantile function for the normal distribution.
 *
 *     For small to moderate probabilities, algorithm referenced
 *     below is used to obtain an initial approximation which is
 *     polished with a final Newton step.
 *
 *     For very large arguments, an algorithm of Wichura is used.
 *
 *  REFERENCE
 *
 *     Beasley, J. D. and S. G. Springer (1977).
 *     Algorithm AS 111: The percentage points of the normal distribution,
 *     Applied Statistics, 26, 118-121.
 *
 *      Wichura, M.J. (1988).
 *      Algorithm AS 241: The Percentage Points of the Normal Distribution.
 *      Applied Statistics, 37, 477-484.
 */
 
function normsInv(p, mu, sigma)
{
    if (p < 0 || p > 1)
    {
        throw "The probality p must be bigger than 0 and smaller than 1";
    }
    if (sigma < 0)
    {
        throw "The standard deviation sigma must be positive";
    }
 
    if (p == 0)
    {
        return -Infinity;
    }
    if (p == 1)
    {
        return Infinity;
    }
    if (sigma == 0)
    {
        return mu;
    }
 
    var q, r, val;
 
    q = p - 0.5;
 
    /*-- use AS 241 --- */
    /* double ppnd16_(double *p, long *ifault)*/
    /*      ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3
 
            Produces the normal deviate Z corresponding to a given lower
            tail area of P; Z is accurate to about 1 part in 10**16.
    */
    if (Math.abs(q) <= .425)
    {/* 0.075 <= p <= 0.925 */
        r = .180625 - q * q;
        val =
               q * (((((((r * 2509.0809287301226727 +
                          33430.575583588128105) * r + 67265.770927008700853) * r +
                        45921.953931549871457) * r + 13731.693765509461125) * r +
                      1971.5909503065514427) * r + 133.14166789178437745) * r +
                    3.387132872796366608)
               / (((((((r * 5226.495278852854561 +
                        28729.085735721942674) * r + 39307.89580009271061) * r +
                      21213.794301586595867) * r + 5394.1960214247511077) * r +
                    687.1870074920579083) * r + 42.313330701600911252) * r + 1);
    }
    else
    { /* closer than 0.075 from {0,1} boundary */
 
        /* r = min(p, 1-p) < 0.075 */
        if (q > 0)
            r = 1 - p;
        else
            r = p;
 
        r = Math.sqrt(-Math.log(r));
        /* r = sqrt(-log(r))  <==>  min(p, 1-p) = exp( - r^2 ) */
 
        if (r <= 5)
        { /* <==> min(p,1-p) >= exp(-25) ~= 1.3888e-11 */
            r += -1.6;
            val = (((((((r * 7.7454501427834140764e-4 +
                       .0227238449892691845833) * r + .24178072517745061177) *
                     r + 1.27045825245236838258) * r +
                    3.64784832476320460504) * r + 5.7694972214606914055) *
                  r + 4.6303378461565452959) * r +
                 1.42343711074968357734)
                / (((((((r *
                         1.05075007164441684324e-9 + 5.475938084995344946e-4) *
                        r + .0151986665636164571966) * r +
                       .14810397642748007459) * r + .68976733498510000455) *
                     r + 1.6763848301838038494) * r +
                    2.05319162663775882187) * r + 1);
        }
        else
        { /* very close to  0 or 1 */
            r += -5;
            val = (((((((r * 2.01033439929228813265e-7 +
                       2.71155556874348757815e-5) * r +
                      .0012426609473880784386) * r + .026532189526576123093) *
                    r + .29656057182850489123) * r +
                   1.7848265399172913358) * r + 5.4637849111641143699) *
                 r + 6.6579046435011037772)
                / (((((((r *
                         2.04426310338993978564e-15 + 1.4215117583164458887e-7) *
                        r + 1.8463183175100546818e-5) * r +
                       7.868691311456132591e-4) * r + .0148753612908506148525)
                     * r + .13692988092273580531) * r +
                    .59983220655588793769) * r + 1);
        }
 
        if (q < 0.0)
        {
            val = -val;
        }
    }
 
    return mu + sigma * val;
}
 
 
