from django.http import HttpResponse, HttpResponseRedirect, HttpRequest
from django.template import RequestContext, loader
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.decorators import login_required
from django.contrib.auth import authenticate, login, logout
from django.contrib.sites.models import Site
from rest_framework.response import Response
from rest_framework import status
from visualization.models import *
import safe.views, csv, json, requests


def index(request):
    return home(request)

#================================
# Pages
#================================

# Page for account information/changes
def account(request):
    userId = request.user.id
    accountForm = AccountForm(request.POST or None, user=request.user)

    if accountForm.is_valid():
        if accountForm.cleaned_data["new_password"]:
            request.user.set_password(accountForm.cleaned_data["new_password"])
        request.user.save()

    t = loader.get_template('visualization/page_account.html')
    c = RequestContext(request, {'accountForm': accountForm})
    return HttpResponse(t.render(c))

# Page for experiment creation
def createExp(request):
    userId = request.user.id
    userId = 1

    descriptions = ExpDescription.objects.filter(owner_id=userId) | ExpDescription.objects.filter(public=1)
    
    expDescriptionSelectForm = ExpDescriptionSelectForm(request.POST or None, expDescs=descriptions)
    expDescriptionInfoForm = None

    # descriptions available for this user
    if len(descriptions) > 0:
        expDescriptionInfoForm = ExpDescriptionInfoForm(expDescription=descriptions[0])
        c = RequestContext(request, {'expDescriptionSelectForm': expDescriptionSelectForm,
                                     'expDescriptionInfoForm': expDescriptionInfoForm })
    # descriptions unavailable for this user
    else:
        c = RequestContext(request, {'expDescriptionSelectForm': expDescriptionSelectForm,
                                     'expDescriptionInfoForm': expDescriptionInfoForm,
                                     'error': 'No available experiment descriptions.' })

    t = loader.get_template('visualization/page_create_experiment.html')
    return HttpResponse(t.render(c))

# Page for creating a new user
def createUser(request):
    createForm = UserCreateForm(request.POST or None)

    if request.method == "POST" and createForm.is_valid():
        first = request.POST["first_name"]
        last = request.POST["last_name"]
        
        username = request.POST["username"]
        password = request.POST["password"]
        email = request.POST["email"]

        newUser = User.objects.create_user(username, email, password)
        newUser.first_name = first
        newUser.last_name = last
        newUser.save()

        user = authenticate(username=username, password=password)
        if user is not None:
            t = loader.get_template('visualization/template_create_user_success.html')
            c = RequestContext(request)
            HttpResponse(t.render(c))
        else:
            t = loader.get_template('visualization/template_create_user_fail.html')
            c = RequestContext(request)
            return HttpResponse(t.render(c))

    t = loader.get_template('visualization/page_create_user.html')
    c = RequestContext(request, {"createForm": createForm})
    return HttpResponse(t.render(c))

# Page holding experiments' statuses
def expStatus(request):
    user_id = request.user.id
    experiments = Experiment.objects.filter(owner_id = user_id) | Experiment.objects.filter(public = 1)

    t = loader.get_template('visualization/page_status.html')
    c = RequestContext(request, {'experiments':experiments})
    return HttpResponse(t.render(c))

# Home page
def home(request):
    t = loader.get_template('visualization/page_home.html')
    c = RequestContext(request, {})
    return HttpResponse(t.render(c));

# Page for downloading experiment results
def results(request):
    userId = request.user.id
    expId = 1
    experiment = Experiment.objects.get(pk=expId)

    t = loader.get_template('visualization/page_results.html')
    c = RequestContext(request, { "experiment": experiment })
    return HttpResponse(t.render(c))

# Page for selecting visualizations to be used
def visualizations(request):
    user_id = request.user.id
    experiments = Experiment.objects.filter(owner_id = user_id) | Experiment.objects.filter(public = 1)

    t = loader.get_template('visualization/page_visualizations.html')
    c = RequestContext(request, {'experiments':experiments})
    return HttpResponse(t.render(c))

#================================
# Page elements
#================================

# The sub-form for experiment creation
def getCreateExp(request):
    # Check for permissions
    if(not request.user.is_authenticated()):
        t = loader.get_template("visualization/template_login_required_error.html")
        c = RequestContext(request, { "message": "Must be logged in to create an experiment." })
        return HttpResponse(json.dumps({"status": "INVALID", "content":t.render(c) }))

    expDescriptionId = request.POST["expDescription"]
    expDescription = ExpDescription.objects.get(pk=expDescriptionId)
    factors = expDescription.factor_set.all()

    # Single argument is the expDescriptionId for initialization
    # Multiple indicate a validation attempt
    if len(request.POST)  > 1:
        expForm = ExpForm(request.POST)
        factorForm = FactorForm(request.POST, factors=factors)

        if expForm.is_valid() and factorForm.is_valid():
            upload = {}

            upload["exp_description_id"] = expDescription.id
            upload["name"] = expForm.cleaned_data["name"]
            upload["description"] = expForm.cleaned_data["description"]
            upload["replications"] = expForm.cleaned_data["replications"]
            upload["public"] = expForm.cleaned_data["public"]
            upload["factors"] = [] 

            for factor, result in factorForm.cleaned_data.values():
                myFactor = {}
                myFactor["name"]        = factor.name
                myFactor["description"] = factor.description
                myFactor["type"]        = factor.data_type
                myFactor["visible"]     = factor.visible
                myFactor["levels"]      = result

                upload["factors"].append(myFactor)

            result = {}
            result["content"] = upload
            result["status"] = "VALID"
            return HttpResponse(json.dumps(result))
    else:
        expForm = ExpForm()
        factorForm = FactorForm(factors=factors)

    t = loader.get_template('visualization/template_create_experiment.html')
    c = RequestContext(request, {'expForm': expForm, 'factorForm': factorForm, 'expDescription': expDescription })

    result = {}
    result["content"] = t.render(c)
    result["status"] = "INVALID"
    return HttpResponse(json.dumps(result))

# Sub-form for user creation
def getCreateUser(request):
    createForm = UserCreateForm(request.POST or None)

    t = loader.get_template('visualization/template_create_user.html')
    c = RequestContext(request, {"createForm": createForm })
    return HttpResponse(t.render(c))

# Sub-form for home page login
def getLogin(request):
    if request.method == "POST":
        loginForm = AuthenticationForm(data=request.POST)
    else:
        loginForm = AuthenticationForm(request)

    if loginForm.is_valid():
        username = request.POST["username"]
        password = request.POST["password"]
        user = authenticate(username=username, password=password)
        if user is not None:
            login(request, user)
            return HttpResponse("OK")

    t = loader.get_template('visualization/template_login.html')
    c = RequestContext(request, {"loginForm": loginForm})
    return HttpResponse(t.render(c))

# Sub-form for drop-down login
def getLoginDropdown(request):
    errors = {'required': 'Required',
              'invalid': 'Invalid'}

    if request.method == "POST":
        loginForm = AuthenticateDropdownForm(data=request.POST, error_messages=errors)
    else:
        loginForm = AuthenticateDropdownForm(request, error_messages=errors)

    if loginForm.is_valid():
        username = request.POST["username"]
        password = request.POST["password"]
        user = authenticate(username=username, password=password)
        if user is not None:
            login(request, user)
            return HttpResponse("OK")

    t = loader.get_template('visualization/template_login_dropdown.html')
    c = RequestContext(request, {"loginForm": loginForm})
    return HttpResponse(t.render(c))

# Sub-form for designpoint selection
def getSelectDesignpoint(request):
    decoder = json.JSONDecoder()
    data = decoder.decode(request.body)

    expId = data["experiment"]

    experiment = Experiment.objects.get(pk=expId)
    designpoints = DesignPoint.objects.filter(experiment=experiment)
    designpointForm = DesignpointForm(designpoints=designpoints) 

    t = loader.get_template('visualization/template_select_designpoint.html')
    c = RequestContext(request, {"designpointSelectForm": designpointForm})
    return HttpResponse(t.render(c))

# Sub-form for experiment selection
def getSelectExp(request):
    userId = request.user.id

    experiments = Experiment.objects.filter(owner_id=userId) | Experiment.objects.filter(public=1)
    expSelectForm = ExpSelectForm(experiments=experiments)

    t = loader.get_template('visualization/template_select_experiment.html')
    c = RequestContext(request, {"expSelectForm": expSelectForm})
    return HttpResponse(t.render(c))

# Sub-form for metric selection
def getSelectMetric(request):
    decoder = json.JSONDecoder()
    data = decoder.decode(request.body)

    expId = data["experiment"]

    experiment = Experiment.objects.get(pk=expId)
    metrics = experiment.exp_description.metric_set.all()
    metricForm = MetricForm(metrics=metrics)

    t = loader.get_template('visualization/template_select_metric.html')
    c = RequestContext(request, {"metricSelectForm": metricForm})
    return HttpResponse(t.render(c))

# Sub-form for run selection
def getSelectRun(request):
    decoder = json.JSONDecoder()
    data = decoder.decode(request.body)

    dpId = data["designpoint"]

    designpoint = DesignPoint.objects.filter(pk=dpId)
    simruns = SimulationRun.objects.filter(design_point=designpoint)
    simrunForm = SimrunForm(simruns=simruns) 

    t = loader.get_template('visualization/template_select_run.html')
    c = RequestContext(request, {"runSelectForm": simrunForm })
    return HttpResponse(t.render(c))

#================================
# Context Actions
#================================

# Perform logout dependent on current user
def logoutAction(request):
    logout(request)
    return HttpResponse("OK")

#================================
# Graphs 
#================================

# Page for graph creation (same format for all graphs)
# @login_required
def genericGraph(request, graph_type):
    #t = loader.get_template('visualization/' + graph_type + '.html')
    t = loader.get_template('visualization/page_generic_graph.html')
    c = RequestContext(request, {'js': '/static/visualization/js/layout/' + graph_type + '.js',
                                 'css': '/static/visualization/css/' + graph_type + '.css'});
    return HttpResponse(t.render(c))

#================================
# General
#================================

def query(request):
    params = request.GET
    keys = params.keys()

    r = None

    if sorted(keys) == [u'metric', u'simrun']:
        r = requests.get(request.build_absolute_uri('/safe/sim-run/'+str(params['simrun'])+'/metric/'+str(params['metric'])+'/results/'))
    elif keys == ['exp_id']:
        r = requests.get(request.build_absolute_uri('/safe/exp-descs/'+str(params['exp_id'])+'/metrics/'))

    else:
        print('lolno')
        return
    out = r.text
    
    return HttpResponse(out, content_type='application/json')
